from itertools import chain
from operator import attrgetter
from typing import Literal, List

from django.db.models import Q
from django.db.models import QuerySet

from drf_stripe.stripe_api.api import stripe_api as stripe
from ..models import Subscription, Price
from ..stripe_models.subscription import StripeSubscriptionStatus

"""
status argument, see https://stripe.com/docs/api/subscriptions/list?lang=python#list_subscriptions-status
"""
STATUS_ARG = Literal[
    "active",
    "past_due",
    "unpaid",
    "canceled",
    "incomplete",
    "incomplete_expired",
    "trialing",
    "all",
    "ended"
]


def stripe_api_list_subscriptions(status: STATUS_ARG = None, limit: int = 100, starting_after: str = None):
    """
    Retrieve all subscriptions.

    """
    stripe.Subscription.list(status=status, limit=limit, starting_after=starting_after)


def list_user_subscriptions(user_id, current=True) -> QuerySet[Subscription]:
    """
    Retrieve a set of Subscriptions associated with a given User instance.

    :param user_id: Django User id.
    :param bool current: Defaults to True and retrieves only current subscriptions
        (excluding any cancelled, ended, unpaid subscriptions)
    """
    q = Q(user_id=user_id)
    if current is True:
        q &= Q(status__in={StripeSubscriptionStatus.ACTIVE, StripeSubscriptionStatus.PAST_DUE,
                           StripeSubscriptionStatus.TRIALING})

    return Subscription.objects.filter(q)


def list_user_subscription_products(user_id, current=True):
    """
    Retrieve a set of Product instances associated with a given User instance.

    :param user_id: Django User id.
    :param bool current: Defaults to True and retrieves only products associated with current subscriptions
        (excluding any cancelled, ended, unpaid subscription products)
    """
    subscriptions = list_user_subscriptions(user_id, current=current)
    sub_items = chain.from_iterable(
        sub.items.all() for sub in subscriptions.all().prefetch_related("items__price__product"))
    products = set(item.price.product for item in sub_items)
    return products


def list_subscribable_product_prices_to_user(user_id):
    """
    Retrieve a set of Price instances associated with Products that the User isn't currently subscribed to.

    :param user_id: Django user id.
    """
    current_products = set(map(attrgetter('product_id'), list_user_subscription_products(user_id)))
    prices = Price.objects.filter(
        Q(active=True) &
        Q(product__active=True) &
        ~Q(product__product_id__in=current_products)
    )
    return prices


def list_all_available_product_prices(expand: List = None):
    """Retrieve a set of all Price instances that are available to public."""

    prices = Price.objects.filter(Q(active=True) & Q(product__active=True))

    if expand and "feature" in expand:
        prices = prices.prefetch_related("product__linked_features__feature")

    return prices



from rest_framework import serializers

from drf_stripe.models import SubscriptionItem, Product, Price, Subscription


class SubscriptionSerializer(serializers.ModelSerializer):
    class Meta:
        model = Subscription
        fields = (
            "subscription_id", "period_start", "period_end", "status", "cancel_at", "cancel_at_period_end",
            "trial_start", "trial_end"
        )


class SubscriptionItemSerializer(serializers.ModelSerializer):
    """Serializes SubscriptionItem model with attributes pulled from related Subscription instance"""
    product_id = serializers.CharField(source="price.product.product_id")
    price_id = serializers.CharField(source="price.price_id")
    price = serializers.CharField(source="price.price")
    freq = serializers.CharField(source="price.freq")
    services = serializers.SerializerMethodField(method_name='get_feature_ids')
    subscription_status = serializers.CharField(source='subscription.status')
    period_start = serializers.DateTimeField(source='subscription.period_start')
    period_end = serializers.DateTimeField(source='subscription.period_end')
    trial_start = serializers.DateTimeField(source='subscription.trial_start')
    trial_end = serializers.DateTimeField(source='subscription.trial_end')
    ended_at = serializers.DateTimeField(source='subscription.ended_at')
    cancel_at = serializers.DateTimeField(source='subscription.cancel_at')
    cancel_at_period_end = serializers.BooleanField(source='subscription.cancel_at_period_end')

    def get_feature_ids(self, obj):
        return [{"feature_id": link.feature.feature_id, "feature_desc": link.feature.description} for link in
                obj.price.product.linked_features.all().prefetch_related('feature')]

    def get_subscription_expires_at(self, obj):
        return obj.subscription.period_end or \
               obj.subscription.cancel_at or \
               obj.subscription.trial_end or \
               obj.subscription.ended_at

    class Meta:
        model = SubscriptionItem
        fields = (
            "product_id", "price_id", "price", "freq", "subscription_status", "period_start", "period_end",
            "trial_start", "trial_end", "ended_at", "cancel_at", "cancel_at_period_end", "services")


class ProductSerializer(serializers.ModelSerializer):
    class Meta:
        model = Product
        fields = "__all__"


class PriceSerializer(serializers.ModelSerializer):
    product_id = serializers.CharField(source="product.product_id")
    name = serializers.CharField(source="product.name")
    avail = serializers.BooleanField(source="active")
    services = serializers.SerializerMethodField(method_name='get_feature_ids')

    def get_feature_ids(self, obj):
        return [{"feature_id": prod_feature.feature.feature_id, "feature_desc": prod_feature.feature.description} for
                prod_feature in
                obj.product.linked_features.all().prefetch_related("feature")]

    class Meta:
        model = Price
        fields = ("price_id", "product_id", "name", "price", "freq", "avail", "services")



from django.contrib.auth import get_user_model
from django.db import models


class StripeUser(models.Model):
    """A model linking Django user model with a Stripe User"""
    user = models.OneToOneField(get_user_model(), on_delete=models.CASCADE, related_name='stripe_user',
                                primary_key=True)
    customer_id = models.CharField(max_length=128, null=True)

    class Meta:
        indexes = [
            models.Index(fields=['user', 'customer_id'])
        ]


class Feature(models.Model):
    """
    A model used to keep track of features provided by your application.
    This does not correspond to a Stripe object, but the feature ids should be listed as
     a space delimited strings in Stripe.product.metadata.features
    """
    feature_id = models.CharField(max_length=64, primary_key=True)
    description = models.CharField(max_length=256, null=True)


class Product(models.Model):
    """A model representing a Stripe Product"""
    product_id = models.CharField(max_length=256, primary_key=True)
    active = models.BooleanField()
    description = models.CharField(max_length=1024, null=True)
    name = models.CharField(max_length=256, null=True)


class ProductFeature(models.Model):
    """A model representing association of Product and Feature instances. They have many-to-many relationship."""
    product = models.ForeignKey(Product, on_delete=models.CASCADE, related_name="linked_features")
    feature = models.ForeignKey(Feature, on_delete=models.CASCADE, related_name="linked_products")


class Price(models.Model):
    """A model representing to a Stripe Price object, with enhanced attributes."""
    price_id = models.CharField(max_length=256, primary_key=True)
    product = models.ForeignKey(Product, on_delete=models.CASCADE, related_name="prices")
    nickname = models.CharField(max_length=256, null=True)  # displayed name
    price = models.PositiveIntegerField()  # price in cents, corresponding to Stripe unit_amount
    # billing frequency, translated from Stripe price.recurring.interval and price.recurring.interval_count
    freq = models.CharField(max_length=64, null=True)
    active = models.BooleanField()

    class Meta:
        indexes = [
            models.Index(fields=['active', 'freq'])
        ]


class Subscription(models.Model):
    """
    A model representing the association of User and Price, corresponding to a Stripe Subscription object single
    line item.
    """
    subscription_id = models.CharField(max_length=256, primary_key=True)
    user = models.ForeignKey(get_user_model(), on_delete=models.CASCADE, related_name="subscriptions")
    period_start = models.DateTimeField(null=True)
    period_end = models.DateTimeField(null=True)
    cancel_at = models.DateTimeField(null=True)
    cancel_at_period_end = models.BooleanField()
    ended_at = models.DateTimeField(null=True)
    status = models.CharField(max_length=64)
    trial_end = models.DateTimeField(null=True)
    trial_start = models.DateTimeField(null=True)

    class Meta:
        indexes = [
            models.Index(fields=['user', 'status'])
        ]


class SubscriptionItem(models.Model):
    sub_item_id = models.CharField(max_length=256, primary_key=True)
    subscription = models.ForeignKey(Subscription, on_delete=models.CASCADE, related_name="items")
    price = models.ForeignKey(Price, on_delete=models.CASCADE, related_name="+")
    quantity = models.PositiveIntegerField()



"""
Events that Webhook may receive as part of subscription life cycle

Initial subscription:
- payment_method.attached *
- setup_intent.succeeded
- setup_intent.created
- checkout.session.completed
- invoice.created
- invoice.finalized
- invoice.paid
- invoice.payment_succeeded
- customer.subscription.created

Renewal payment declined:
- charge.failed
- invoice.payment_failed
- payment_intent.created
- payment_intent.payment_failed
- payment_intent.canceled

Renewal payment successful:
- charge.succeeded
- customer.subscription.trial_will_end *
- invoice.created
- invoice.finalized
- invoice.paid
- invoice.payment_succeeded
- customer.subscription.updated
- payment_intent.succeeded
- payment_intent.created

Subscription cancellation at period end
- customer.subscription.updated

Subscription immediate cancellation
- customer.subscription.deleted

"""

from django.contrib.auth import get_user_model

from drf_stripe.models import Subscription, SubscriptionItem
from drf_stripe.stripe_models.event import StripeSubscriptionEventData


def _handle_customer_subscription_event_data(data: StripeSubscriptionEventData):
    subscription_id = data.object.id
    customer = data.object.customer
    period_start = data.object.current_period_start
    period_end = data.object.current_period_end
    cancel_at_period_end = data.object.cancel_at_period_end
    cancel_at = data.object.cancel_at
    ended_at = data.object.ended_at
    status = data.object.status
    trial_end = data.object.trial_end
    trial_start = data.object.trial_start
    items = data.object.items

    user = get_user_model().objects.get(stripe_user__customer_id=customer)

    subscription, created = Subscription.objects.update_or_create(
        subscription_id=subscription_id,
        defaults={
            "user": user,
            "period_start": period_start,
            "period_end": period_end,
            "cancel_at": cancel_at,
            "cancel_at_period_end": cancel_at_period_end,
            "ended_at": ended_at,
            "status": status,
            "trial_end": trial_end,
            "trial_start": trial_start
        })

    print(
        f"{subscription} created={created} status={status}, cancel_at={cancel_at}, cancel_at_period_end={cancel_at_period_end}, trial_end={trial_end}, period_start={period_start}, period_end={period_end}")

    for item in items.data:
        item_id = item.id
        price = item.price.id
        quantity = item.quantity
        SubscriptionItem.objects.update_or_create(
            sub_item_id=item_id,
            defaults={
                "subscription": subscription,
                "price_id": price,
                "quantity": quantity
            }
        )



from typing import overload

from django.contrib.auth import get_user_model

from drf_stripe.models import StripeUser
from drf_stripe.stripe_api.api import stripe_api as stripe


@overload
def get_or_create_stripe_user(user_instance) -> StripeUser:
    ...


@overload
def get_or_create_stripe_user(user_id, user_email) -> StripeUser:
    ...


@overload
def get_or_create_stripe_user(user_id) -> StripeUser:
    ...


def get_or_create_stripe_user(**kwargs) -> StripeUser:
    """
    Get or create a StripeUser given a User instance, or given user id and user email.

    :key user_instance: Django user instance.
    :key user_id: Django User id.
    :key user_email: user email address
    """
    user_instance = kwargs.get("user_instance")
    user_id = kwargs.get("user_id")
    user_email = kwargs.get("user_email")

    if user_instance and isinstance(user_instance, get_user_model()):
        return _get_or_create_stripe_user_from_user_instance(user_instance)
    elif user_id and user_email and isinstance(user_id, str):
        return _get_or_create_stripe_user_from_user_id_email(user_id, user_email)
    elif user_id is not None:
        return _get_or_create_stripe_user_from_user_id(user_id)
    else:
        raise TypeError("Unknown keyword arguments!")


def _get_or_create_stripe_user_from_user_instance(user_instance):
    """
    Return a StripeUser instance given a Django User instance.

    :param user_instance: Django User instance.
    """
    return _get_or_create_stripe_user_from_user_id_email(user_instance.id, user_instance.email)


def _get_or_create_stripe_user_from_user_id(user_id):
    """
    Return a StripeUser instance given user_id.

    :param user_id: user id
    """
    user = get_user_model().objects.get(id=user_id)

    return _get_or_create_stripe_user_from_user_id_email(user.id, user.email)


def _get_or_create_stripe_user_from_user_id_email(user_id, user_email: str):
    """
    Return a StripeUser instance given user_id and user_email.

    :param user_id: user id
    :param string user_email: user email address
    """
    stripe_user, created = StripeUser.objects.get_or_create(user_id=user_id)

    if created:
        customer = stripe.Customer.create(email=user_email)
        stripe_user.customer_id = customer.id
        stripe_user.save()

    return stripe_user
